"""
    Main app components definition.
"""
import signal
import sys

from pkg_resources import get_distribution
from cliff.app import App
from cliff.commandmanager import CommandManager

from digicloud.cli import signal_handler
from .managers import ConfigManager, Session


class DigicloudApp(App):
    """Overwrite ``cliff.app.App`` class.

    Make possibility to further extension and overwrite methods.
    """

    def __init__(self):
        command_manager = CommandManager('digicloud.cli')
        super(DigicloudApp, self).__init__(
            description='Digicloud CLI',
            version=get_distribution('digicloud').version,
            command_manager=command_manager,
            deferred_help=True
        )
        self.config = ConfigManager(self.LOG)
        # Http session from requests.Session.
        base_url = self.config['BASE_URL']
        self.session = self.config.get('SESSION', Session(base_url))

    def initialize_app(self, argv):
        signal.signal(signal.SIGINT, signal_handler.interrupt_handler)
        self.LOG.debug('Initialize session')
        self.session.base_url = self.config['BASE_URL']

        if self.config.get('AUTH_HEADERS'):
            auth_headers = self.config['AUTH_HEADERS']
            self.session.headers.update(auth_headers)
            self.config['SESSION'] = self.session
        else:
            self.config['AUTH_HEADERS'] = {}

    def clean_up(self, cmd, result, err):
        """Is invoked after a command runs."""
        self.config.save()
        self.LOG.debug('Configuration saved.')
        if err:
            self.LOG.debug(err)


def main(argv=None):
    """Initialize main ``cliff.app.App`` instance and run.

    Cliff look for this function as a console script entry point.
    """
    if not argv:
        argv = sys.argv[1:]
    if len(argv) == 0:  # Disable interactive mode
        argv = ['--help']  # display --help instead of interactive mode
    return DigicloudApp().run(argv)


if __name__ == '__main__':
    sys.exit(main())



from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from digicloud.utils import tabulate


class ListRegions(Lister):
    """List available regions"""
    def take_action(self, parsed_args):
        data = self.app.session.get('/regions')
        return tabulate(data)


class SelectRegion(Command):
    """Select specific region to setup"""
    def get_parser(self, prog_name):
        parser = super(SelectRegion, self).get_parser(prog_name)

        parser.add_argument(
            'region_name',
            metavar='<region_name>',
            help="Region name"
        )

        return parser

    def take_action(self, parsed_args):
        self.app.config['AUTH_HEADERS'].update({
            'Digicloud-Region': parsed_args.region_name,
        })


class CurrentRegion(ShowOne):
    """Show current selected region"""
    def take_action(self, parsed_args):
        header = self.app.config.get('AUTH_HEADERS')
        if header is not None:
            selected_region = header.get('Digicloud-Region', 'No region selected')
        else:
            selected_region = 'No region selected'

        return ('Current region', ), (selected_region, )



"""
    DigiCloud InstanceTypes Service.
"""
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate


class ListInstanceType(Lister):
    """List instance-types."""

    def take_action(self, parsed_args):
        uri = '/instance-types'
        data = self.app.session.get(uri)

        return tabulate(data)


class ShowInstanceType(ShowOne):
    """Show instance-type details."""

    def get_parser(self, prog_name):
        parser = super(ShowInstanceType, self).get_parser(prog_name)
        parser.add_argument(
            'instance_type',
            metavar='<instance type>',
            help='InstanceType name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = f'/instance-types/{parsed_args.instance_type}'
        data = self.app.session.get(uri)

        return tabulate(data)



"""
    DigiCloud Network Service.
"""
from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate


class ListNetwork(Lister):
    """List networks"""

    def take_action(self, parsed_args):
        data = self.app.session.get('/networks')

        return tabulate(data)


class ShowNetwork(ShowOne):
    """Show network details."""

    def get_parser(self, prog_name):
        parser = super(ShowNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        data = self.app.session.get(uri)

        return tabulate(data)


class DeleteNetwork(Command):
    """Delete network."""

    def get_parser(self, prog_name):
        parser = super(DeleteNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        self.app.session.delete(uri)


class UpdateNetwork(ShowOne):
    """Update network."""
    def get_parser(self, prog_name):
        parser = super(UpdateNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network ID',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for network.'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<Description>',
            help='New admin state.'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/networks/%s' % parsed_args.network
        payload = {k: getattr(parsed_args, k) for k in ('name', 'admin_state')
                   if getattr(parsed_args, k) is not None}
        data = self.app.session.patch(uri, payload)
        return tabulate(data)


class CreateNetwork(ShowOne):
    """Create Network"""

    def get_parser(self, prog_name):
        parser = super(CreateNetwork, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Network name'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            default='UP',
            help='Set admin state (default UP).'
        )
        return parser

    def take_action(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'admin_state': parsed_args.admin_state
        }
        data = self.app.session.post('/networks', payload)
        return tabulate(data)


class ListNetworkPorts(Lister):
    """List all ports of Network"""
    def get_parser(self, prog_name):
        parser = super(ListNetworkPorts, self).get_parser(prog_name)

        parser.add_argument(
            'network',
            metavar='<network>',
            help='Network name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = f'/networks/{parsed_args.network}/ports'
        data = self.app.session.get(uri)

        return tabulate(data)



"""
    Authentication to digicloud API.
"""
import getpass

from cliff.show import ShowOne
from cliff.command import Command


class Login(ShowOne):
    """Digicloud Login"""

    def get_parser(self, prog_name):
        parser = super(Login, self).get_parser(prog_name)
        parser.add_argument(
            '--email',
            required=True,
            metavar='<EMAIL>',
            dest='email',
            help=("Your Digicloud email")
        )

        parser.add_argument(
            '--password',
            metavar='<PASSWORD>',
            required=False,
            default=None,
            dest='password',
            help=("Your Digicloud account password")
        )

        parser.add_argument(
            '--namespace',
            metavar='<NAMESPACE>',
            dest='namespace',
            help=("ID of the Digicloud namespace")
        )

        return parser

    def take_action(self, parsed_args):
        columns = ('Email', 'Namespace name', 'Namespace ID',)
        payload = {
            'email': parsed_args.email,
            'password': parsed_args.password or getpass.getpass(),
        }
        data = self.app.session.post('/tokens', payload)

        selected_ns = None
        if parsed_args.namespace:
            selected_ns = self._find_namespace(parsed_args.namespace, data['namespaces'])

        selected_ns = selected_ns or data['namespaces'][0]
        rows = (parsed_args.email, selected_ns['name'], selected_ns['id'])
        self.app.config['AUTH_HEADERS'].update({
            'Authorization': f'Bearer {data["token"]}',
            'Digicloud-Namespace': selected_ns['id'],
        })
        self.app.config['USER'] = data['user']
        return columns, rows

    def _find_namespace(self, name_or_id, namespaces):
        result = [ns for ns in namespaces
                  if name_or_id == ns['name'] or name_or_id == ns['id']]
        if len(result) == 0:
            raise Exception(f'No namespace matches {name_or_id}')
        elif len(result) > 1:
            raise Exception(f'Multitple namespaces match {name_or_id}')
        return result[0]


class Logout(Command):
    """DigiCloud Logout"""

    def take_action(self, parsed_args):
        del self.app.config['AUTH_HEADERS']
        del self.app.session.headers['Authorization']

        return 0





"""
    DigiCloud Access Service.
"""
from cliff.lister import Lister
from cliff.show import ShowOne
from cliff.command import Command

from ..utils import tabulate


class ListFloatingIp(Lister):
    """List floating Ips."""

    def take_action(self, parsed_args):
        uri = '/floating-ips'
        data = self.app.session.get(uri)

        return tabulate(data)


class ShowFloatingIp(ShowOne):
    """Show floating Ip details."""

    def get_parser(self, prog_name):
        parser = super(ShowFloatingIp, self).get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating_ip>',
            help=('Floating IP ID'),
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/floating-ips/%s' % parsed_args.floating_ip
        data = self.app.session.get(uri)

        return tabulate(data)


class CreateFloatingIp(ShowOne):
    """Create Floating Ip."""
    def take_action(self, parsed_args):
        data = self.app.session.post('/floating-ips', {})

        return tabulate(data)


class DeleteFloatingIp(Command):
    """Delete Floating Ip."""

    def get_parser(self, prog_name):
        parser = super(DeleteFloatingIp, self).get_parser(prog_name)
        parser.add_argument(
            'floating_ip',
            metavar='<floating_ip>',
            help=('Floating IP ID')
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/floating-ips/%s' % parsed_args.floating_ip
        self.app.session.delete(uri)


class AssociateFloatingIp(ShowOne):
    """Associate Floating Ip."""

    def get_parser(self, prog_name):
        parser = super(AssociateFloatingIp, self).get_parser(prog_name)
        parser.add_argument(
            'floating_ip_id',
            metavar='<floating_ip_id>',
            help='Floating IP ID'
        )
        parser.add_argument(
            '--port-id',
            metavar='<port_id>',
            required=True,
            help='Port ID'
        )
        return parser

    def take_action(self, parsed_args):
        payload = {
            'port_id': parsed_args.port_id,
        }
        uri = '/floating-ips/%s/associate' % parsed_args.floating_ip_id
        data = self.app.session.post(uri, payload)
        return tabulate(data)


class RevokeFloatingIp(ShowOne):
    """Revoke Floating Ip."""

    def get_parser(self, prog_name):
        parser = super(RevokeFloatingIp, self).get_parser(prog_name)
        parser.add_argument(
            'floating_ip_id',
            metavar='<floating_ip_id>',
            help='Floating IP ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/floating-ips/%s/revoke' % parsed_args.floating_ip_id
        data = self.app.session.post(uri, {})
        return tabulate(data)



"""
    DigiCloud Router Service.
"""
from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate


class ListRouter(Lister):
    """List Routers."""

    def take_action(self, parsed_args):
        data = self.app.session.get('/routers')

        return tabulate(data)


class ShowRouter(ShowOne):
    """Show Router details."""

    def get_parser(self, prog_name):
        parser = super(ShowRouter, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        data = self.app.session.get(uri)

        return tabulate(data)


class DeleteRouter(Command):
    """Delete router."""

    def get_parser(self, prog_name):
        parser = super(DeleteRouter, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        self.app.session.delete(uri)


class CreateRouter(ShowOne):
    """Create Router."""

    def get_parser(self, prog_name):
        parser = super(CreateRouter, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Router name'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            default='UP',
            help='Set admin state (default UP).'
        )
        return parser

    def take_action(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'admin_state': parsed_args.admin_state
        }
        data = self.app.session.post('/routers', payload)
        return tabulate(data)


class ListRouterInterface(Lister):
    """List all router interfaces"""

    def get_parser(self, prog_name):
        parser = super(ListRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = f'/routers/{parsed_args.router}/interfaces'
        data = self.app.session.get(uri)
        return tabulate(data)


class AddRouterInterface(ShowOne):
    """Add router interface."""

    def get_parser(self, prog_name):
        parser = super(AddRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--subnet',
            required=True,
            metavar='<subnet>',
            help='Subnet name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/interfaces' % parsed_args.router
        payload = {'subnet': parsed_args.subnet}
        data = self.app.session.post(uri, payload)

        return tabulate(data)


class RemoveRouterInterface(ShowOne):
    """Remove router interface."""
    def get_parser(self, prog_name):
        parser = super(RemoveRouterInterface, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--subnet',
            required=True,
            metavar='<subnet>',
            help='Subnet name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/interfaces/%s' % (parsed_args.router, parsed_args.subnet)
        data = self.app.session.delete(uri)
        return tabulate(data)


class AddRouterExternal(ShowOne):
    """Add external network to router."""

    def get_parser(self, prog_name):
        parser = super(AddRouterExternal, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        payload = {'has_gateway': True}
        data = self.app.session.patch(uri, payload)
        return tabulate(data)


class RemoveRouterExternal(ShowOne):
    """Remove external network from router."""

    def get_parser(self, prog_name):
        parser = super(RemoveRouterExternal, self).get_parser(prog_name)
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s' % parsed_args.router
        payload = {'has_gateway': False}
        data = self.app.session.patch(uri, payload)
        return tabulate(data)


class UpdateRouter(ShowOne):
    _description = "Update Router"

    def get_parser(self, prog_name):
        parser = super(UpdateRouter, self).get_parser(prog_name)
        group = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'router',
            metavar='<router>',
            help='Router name or ID'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='New name for the router'
        )
        parser.add_argument(
            '--admin-state',
            metavar='<AdminState>',
            help='Set admin state.'
        )
        group.add_argument(
            '--disable-gateway',
            help='Disconnect router from external network',
            action='store_true'

        )
        group.add_argument(
            '--enable-gateway',
            help='Connect router to external network',
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.disable_gateway:
            payload['has_gateway'] = False
        if parsed_args.enable_gateway:
            payload['has_gateway'] = True
        if parsed_args.admin_state:
            payload['admin_state'] = parsed_args.admin_state
        if len(payload) == 0:
            raise Exception("You need to at least provide one of --name, --enable-gateway or --disable-gateway")
        uri = '/routers/%s' % parsed_args.router
        data = self.app.session.patch(uri, payload)
        return tabulate(data)


class ListRouterStatic(Lister):
    """List router static routes."""
    def get_parser(self, prog_name):
        parser = super(ListRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        data = self.app.session.get(uri)
        return tabulate(data)


class AddRouterStatic(Lister):
    """Add router static routes."""
    def get_parser(self, prog_name):
        parser = super(AddRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        parser.add_argument(
            '--destination',
            metavar='<Network>',
            required=True,
            help='Destination network.'
        )
        parser.add_argument(
            '--nexthop',
            metavar='<IP address>',
            required=True,
            help='Next hop IP address.'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        payload = {
            "destination": parsed_args.destination,
            "nexthop": parsed_args.nexthop,

        }
        data = self.app.session.post(uri, payload)
        return tabulate(data)


class DeleteRouterStatic(Lister):
    """Delete static routes."""
    def get_parser(self, prog_name):
        parser = super(DeleteRouterStatic, self).get_parser(prog_name)
        parser.add_argument(
            'router_id',
            metavar='<router_id>',
            help='Router ID'
        )
        parser.add_argument(
            '--destination',
            metavar='<Network>',
            required=True,
            help='Destination network.'
        )
        parser.add_argument(
            '--nexthop',
            metavar='<IP address>',
            required=True,
            help='Next hop IP address.'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/routers/%s/static-routes' % parsed_args.router_id
        payload = {
            "destination": parsed_args.destination,
            "nexthop": parsed_args.nexthop,

        }
        data = self.app.session.delete(uri, payload)
        return tabulate(data)



"""
    Managers definition.
"""
import os
import pickle
import sys
from pathlib import Path
from collections import ChainMap
from json import JSONDecodeError

import yaml
from requests import Session as Sess, HTTPError


class ConfigManager():
    """Configuration manager for all configs."""
    FILE_NAME = 'config'
    DEFAULTS = {
        'BASE_URL': os.environ.get('API_BASE_URL', "https://api.digicloud.ir"),
    }

    def __init__(self, app_logger):
        self._log = app_logger
        self._path = self._get_config_path()
        self._states = dict()
        self._load()

    def __getitem__(self, item):
        try:
            return ChainMap(self.DEFAULTS, self._states).get(item)
        except KeyError:
            raise f'No "{item}" in configuration.'

    def __setitem__(self, item, value):
        self._states[item] = value

    def __delitem__(self, item):
        self._states[item] = None

    def __call__(self):
        return self._states

    def _get_config_path(self):
        """Create config file directory.

        Handles platform dependent path creation for ``FILE_NAME ``,
        and return full path to config file.
        """
        path = {
            'darwin': '$HOME/Library/Application Support/digicloud',
            'win': '%LOCALAPPDATA%/digicloud',
            'linux': '$HOME/.config/digicloud',
        }[sys.platform]

        path = os.path.expandvars(path)

        if not os.path.isdir(path):
            self._log.debug('No config director found.')
            try:
                os.mkdir(path, mode=0o755)
                self._log.debug(f'Config dir made in "{path}".')
            except IOError:
                raise "Create config path failes."

        full_path = os.path.join(path, self.FILE_NAME)
        Path(full_path).touch()

        return full_path

    def _load(self):
        with open(self._path, 'rb') as file_:
            if os.path.getsize(self._path) > 0:
                self._states = pickle.load(file_)

    def _dump(self):
        with open(self._path, 'wb') as file_:
            pickle.dump(self._states, file_, pickle.HIGHEST_PROTOCOL)

    def get(self, item, default=None):
        """return item from config or default."""
        return self.__getitem__(item) or default

    def save(self):
        """write configurations to file."""
        self._dump()


class Session(Sess):
    """Custom ``requests.Session`` object model."""

    def __init__(self, base_url, *args, **kwargs):
        self.base_url = base_url
        super(Session, self).__init__(*args, **kwargs)

    def resource(self, uri, method='GET', payload=None, params=None, endpoint_version='/v1'):
        """Resource access helper."""
        url = f'{self.base_url}{endpoint_version}{uri}'
        response = self.request(method, url, params=params, json=payload)

        try:
            response.raise_for_status()
            data = response.json()
        except JSONDecodeError:
            # Handel endpoints with none json payload.
            data = {}
        except HTTPError:
            if response.status_code == 401 and not response.url.endswith('tokens'):
                error_msg = 'digicloud: Must login.'
            elif response.headers.get('Content-Type') == 'application/json':
                error_msg = yaml.safe_dump(response.json())
            else:
                error_msg = f'digicloud: {response.reason}({response.status_code})'
            raise Exception(error_msg)

        return data

    def get(self, uri, params=None):
        return self.resource(uri, params=params)

    def post(self, uri, payload):
        return self.resource(uri, 'POST', payload)

    def put(self, uri, payload):
        return self.resource(uri, 'PUT', payload)

    def patch(self, uri, payload):
        return self.resource(uri, 'PATCH', payload)

    def delete(self, uri, payload=None):
        return self.resource(uri, 'DELETE', payload)



"""
    Digicloud Namespace Management.
"""
import sys

from cliff.formatters.table import TableFormatter
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate, yes_or_no


class NamespaceFetchMixin:
    def _build_namespace_list(self):
        namespaces = self.app.session.get('/namespaces')
        ids, names = {}, {}
        for ns in namespaces:
            if ns['name'] not in names:
                names[ns['name']] = []
            names[ns['name']].append(ns)
            ids[ns['id']] = ns
        return ids, names

    def _get_namespace_or_exit(self, namespace_ref, parsed_args):
        ids, names = self._build_namespace_list()
        if namespace_ref in ids:
            return ids[namespace_ref]
        if namespace_ref in names:
            possible_options = names[namespace_ref]
            if len(possible_options) == 1:
                return possible_options[0]
            self._display_options(possible_options, parsed_args)
        else:
            raise Exception("You don't have a namespace with this name or id")

    def _display_options(self, options, parsed_args):
        header, rows = tabulate(options)
        TableFormatter().emit_list(
            header,
            rows,
            self.app.stdout,
            parsed_args
        )
        message = "You have more than one namespaces with this name, " \
                  "please use namespace id"
        raise Exception(message)


class CreateNamespace(ShowOne):
    """Create Namespace."""

    def get_parser(self, prog_name):
        parser = super(CreateNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=('Name of namespace')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=('Description of namespace')
        )
        return parser

    def take_action(self, parsed_args):
        payload = {
            'name': parsed_args.name,
            'description': parsed_args.description,
        }
        data = self.app.session.post('/namespaces', payload)

        return tabulate(data)


class ListNamespace(Lister):
    """List namespaces"""

    def take_action(self, parsed_args):
        data = self.app.session.get('/namespaces')

        return tabulate(data)


class ShowNamespace(ShowOne, NamespaceFetchMixin):
    """Show server details."""

    def get_parser(self, prog_name):
        parser = super(ShowNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def take_action(self, parsed_args):
        return tabulate(self._get_namespace_or_exit(parsed_args.namespace, parsed_args))


class ListNamespaceMember(Lister, NamespaceFetchMixin):
    """List namespace members."""

    def get_parser(self, prog_name):
        parser = super(ListNamespaceMember, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def take_action(self, parsed_args):
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)
        uri = '/namespaces/%s/members' % namespace['id']
        data = self.app.session.get(uri)

        return tabulate(data)


class InviteMember(ShowOne, NamespaceFetchMixin):
    """Add Member to Namespace."""

    def get_parser(self, prog_name):
        parser = super(InviteMember, self).get_parser(prog_name)
        parser.add_argument(
            '--email',
            metavar='<user_email>',
            required=True,
            help='E-mail Address to send the invitation'
        )

        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        return parser

    def take_action(self, parsed_args):
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)
        uri = '/invitations'
        payload = {
            'email': parsed_args.email,
            'namespace': namespace['id']
        }
        data = self.app.session.post(uri, payload)
        return tabulate(data)


class DeleteNamespaceMember(ShowOne, NamespaceFetchMixin):
    """Delete Member from Namespace."""

    def get_parser(self, prog_name):
        parser = super(DeleteNamespaceMember, self).get_parser(prog_name)
        parser.add_argument(
            '--user-id',
            metavar='<user_id>',
            required=True,
            help='ID of user'
        )
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )

        return parser

    def take_action(self, parsed_args):
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)
        uri = f'/namespaces/{namespace["id"]}/members/{parsed_args.user_id}'
        data = self.app.session.delete(uri)

        return tabulate(data)


class LeaveNamespace(ShowOne, NamespaceFetchMixin):
    """Leave a namespace """
    def get_parser(self, prog_name):
        parser = super(LeaveNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )

        return parser

    def take_action(self, parsed_args):
        current_user = self.app.config['USER']['id']
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)
        uri = '/namespaces/%s/members/%s' % (namespace['id'], current_user)
        self.app.session.delete(uri)
        return tabulate(namespace)


class DeleteNamespace(ShowOne, NamespaceFetchMixin):
    """Delete Namespace."""

    def get_parser(self, prog_name):
        parser = super(DeleteNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)

        if not parsed_args.i_am_sure and not self.confirm(namespace, parsed_args):
            self.app.stdout.write("Namespace deletion canceled by user\n")
            sys.exit(0)

        uri = '/namespaces/%s?force=true' % namespace['id']
        self.app.session.delete(uri)
        self.app.stdout.write("Namespace has been deleted\n")
        return tabulate(namespace)

    def confirm(self, namespace, parsed_args):
        uri = '/namespaces/%s/resources' % namespace['id']
        data = self.app.session.get(uri)
        headers = ("region", "type", "name", "description", "spec", "created_at")
        rows = []
        for region in data:
            for instance in region['resources']['instances']:
                rows.append(
                    (
                        region['region_name'],
                        "instance",
                        instance['name'],
                        instance.get('description', ""),
                        "type=%s,status=%s" % (instance['type'], instance['status']),
                        instance['created_at'],
                    )
                )
            for volume in region['resources']['volumes']:
                rows.append(
                    (
                        region['region_name'],
                        "volume",
                        volume['name'],
                        volume.get('description', ""),
                        "size=%sGB,status=%s" % (volume['size'], volume['status']),
                        volume['created_at'],
                    )
                )
            for network in region['resources']['networks']:
                rows.append(
                    (
                        region['region_name'],
                        "network",
                        network['name'],
                        network.get('description', ""),
                        "subnets=%s,status=%s" % (
                            len(network['subnets']), network['status']),
                        network['created_at'],
                    )
                )

            for subnet in region['resources']['subnets']:
                rows.append(
                    (
                        region['region_name'],
                        "subnet",
                        subnet['name'],
                        subnet.get('description', ""),
                        "gateway=%s,cidr=%s" % (
                            subnet['gateway_ip'], subnet['cidr']),
                        subnet['created_at'],
                    )
                )

            for router in region['resources']['routers']:
                rows.append(
                    (
                        region['region_name'],
                        "router",
                        router['name'],
                        router.get('description', ""),
                        "status=%s" % router['status'],
                        router['created_at'],
                    )
                )
            for fip in region['resources']['floating_ips']:
                rows.append(
                    (
                        region['region_name'],
                        "floating-ip",
                        " --- ",
                        fip.get('description', ""),
                        "address=%s,status=%s" % (
                        fip['floating_ip_address'], fip['status']),
                        fip['created_at'],
                    )
                )

            for sg in region['resources']['security_groups']:
                rows.append(
                    (
                        region['region_name'],
                        "security-group",
                        sg['name'],
                        sg.get('description', ""),
                        "rules=%s" % len(sg['security_group_rules']),
                        sg['created_at'],
                    )
                )

        TableFormatter().emit_list(
            headers,
            rows,
            self.app.stdout,
            parsed_args
        )
        return yes_or_no(
            "You're about to delete namespace '%s' with all resources including"
            " Instances, Volumes, Networks. \nPlease review this list "
            "carefully before deleting your namespace, All your data and network"
            " configuration will be lost.\n"
            "Are you sure?" % namespace['name'])


class SelectNamespace(ShowOne, NamespaceFetchMixin):
    """Select Namespace."""

    def get_parser(self, prog_name):
        parser = super(SelectNamespace, self).get_parser(prog_name)
        parser.add_argument(
            'namespace',
            metavar='<namespace>',
            help='Namespace name or id',
        )

        return parser

    def take_action(self, parsed_args):
        namespace = self._get_namespace_or_exit(parsed_args.namespace, parsed_args)
        self.app.config['AUTH_HEADERS']['Digicloud-Namespace'] = namespace['id']
        self.app.config.save()
        return tabulate(namespace)


class CurrentNamespace(ShowOne):
    """Display current namespace."""

    def get_parser(self, prog_name):
        parser = super(CurrentNamespace, self).get_parser(prog_name)

        return parser

    def take_action(self, parsed_args):
        if self.app.config.get('AUTH_HEADERS') is None:
            raise Exception("Login is required")
        else:
            namespace_id = self.app.config['AUTH_HEADERS']['Digicloud-Namespace']
            namespace = self.app.session.get('/namespaces/%s' % namespace_id)
            return tabulate(namespace)


class AcceptInvitation(ShowOne, NamespaceFetchMixin):
    """Accept an invitation to join a namespace"""

    def get_parser(self, prog_name):
        parser = super(AcceptInvitation, self).get_parser(prog_name)
        parser.add_argument(
            'invitation',
            metavar='<invitation>',
            help='Invitation ID to accept or reject'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/invitations/%s' % parsed_args.invitation
        payload = {
            'status': "accepted"
        }
        data = self.app.session.patch(uri, payload)
        return tabulate(data)


class ListInvitation(Lister, NamespaceFetchMixin):
    """List your namespace invitations"""

    def take_action(self, parsed_args):
        uri = '/invitations'
        data = self.app.session.get(uri)
        return tabulate(data)



"""
    DigiCloud Compute Instance Service.
"""

from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate, is_tty


class ListInstance(Lister):
    """List instances."""

    def get_parser(self, prog_name):
        parser = super(ListInstance, self).get_parser(prog_name)
        parser.add_argument(
            '--simple',
            help='Advanced instances',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--advanced',
            help='Simple instances',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        args = (parsed_args.simple, parsed_args.advanced)
        if all(args):
            raise Exception("You need to specify one of --simple or --advanced")

        query_params = {}
        if parsed_args.simple:
            query_params['type'] = 'simple'
        elif parsed_args.advanced:
            query_params['type'] = 'advanced'

        data = self.app.session.get( '/instances', params=query_params)
        return tabulate(data)


class CreateInstance(ShowOne):
    """Create instance."""

    def get_parser(self, prog_name):
        parser = super(CreateInstance, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Instance name'
        )
        parser.add_argument(
            '--instance-type',
            required=True,
            metavar='<instance_type>',
            help='InstanceType name or ID'
        )
        parser.add_argument(
            '--image',
            required=True,
            metavar='<image>',
            help='Image name or ID'
        )
        parser.add_argument(
            '--network',
            required=False,
            metavar='<network>',
            help='Network name or ID'
        )
        parser.add_argument(
            '--ssh-key',
            metavar='<ssh_key>',
            help='SSH key name'
        )

        parser.add_argument(
            '--security-group',
            metavar='<security_group>',
            default=[],
            action='append',
            help='Security group name or ID'
        )
        parser.add_argument(
            '--simple',
            help='Create a instance quickly',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--advanced',
            help='Create a instance with more configuration parameters',
            default=None,
            action='store_true'
        )

        return parser

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                all((parsed_args.advanced, parsed_args.simple)),
                "--advanced and --simple should not be used together",
            ),
            (
                parsed_args.simple and parsed_args.network is not None,
                "--simple and --network should not be used together",
            ),
            (
                parsed_args.simple and parsed_args.security_group,
                "--simple and --security-group should not be used together",
            ),
            (
                parsed_args.advanced and parsed_args.network is None,
                "--advanced requires --network to be present",
            ),
            (
                not parsed_args.simple and parsed_args.network is None,
                "--network is mandatory in advanced mode",
            ),

        ]

        for is_invalid, err_msg in rules:
            if is_invalid:
                raise Exception(err_msg)

    def take_action(self, parsed_args):
        self._check_arg_validity(parsed_args)
        payload = {
            'name': parsed_args.name,
            'instance_type': parsed_args.instance_type,
            'image': parsed_args.image,
            'type': 'simple' if parsed_args.simple else 'advanced',
        }
        if parsed_args.network:
            payload['network'] = parsed_args.network
        if parsed_args.ssh_key:
            payload['ssh_key_name'] = parsed_args.ssh_key
        if parsed_args.security_group:
            payload['security_groups'] = parsed_args.security_group

        data = self.app.session.post('/instances', payload)

        return tabulate(data)


class ShowInstance(ShowOne):
    """Show instance details."""

    def get_parser(self, prog_name):
        parser = super(ShowInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s' % parsed_args.instance
        data = self.app.session.get(uri)

        return tabulate(data)


class DeleteInstance(Command):
    """Delete a instance."""

    def get_parser(self, prog_name):
        parser = super(DeleteInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        parser.add_argument(
            '--delete-volumes',
            metavar='<instance>',
            help='Instance name or ID',
        )

        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        if not self.confirm(parsed_args):
            return
        if parsed_args.delete_volumes:
            volumes_uri = '/instances/%s/volumes' % parsed_args.instance
            volumes = self.app.session.get(volumes_uri)
            for volume in volumes:
                del_vol_uri = '/instances/%s/volumes/%s' % parsed_args.instance, volume['id']
                self.app.session.delete(del_vol_uri)

        uri = '/instances/%s' % parsed_args.instance
        self.app.session.delete(uri)

    def confirm(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            instance = self.app.session.get('/instances/%s' % parsed_args.instance)
            user_response = input(
                "You're about to delete the instance named {}  "
                "Are you sure? [yes/no]".format(
                    instance['name']
                ))
            if user_response == "yes":
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write("Unable to perform 'instance delete' operation in non-interactive mode,"
                                  " without '--i-am-sure' switch\n")
            return False


class StartInstance(Command):
    """Start instance."""

    def get_parser(self, prog_name):
        parser = super(StartInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="ACTIVE"))


class StopInstance(Command):
    """Stop instance."""

    def get_parser(self, prog_name):
        parser = super(StopInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="STOPPED"))


class SuspendInstance(Command):
    """Suspend instance."""

    def get_parser(self, prog_name):
        parser = super(SuspendInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="SUSPEND"))


class ResumeInstance(Command):
    """Resume instance."""

    def get_parser(self, prog_name):
        parser = super(ResumeInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="ACTIVE"))


class RebootInstance(Command):
    """Reboot instance."""

    def get_parser(self, prog_name):
        parser = super(RebootInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )

        parser.add_argument(
            '--type',
            default='SOFT',
            metavar='<type>',
            help='either `SOFT` for a software-level reboot, or `HARD` for a virtual '
                 'power cycle hard reboot',
            choices=['SOFT', 'HARD']
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/state-transitions' % parsed_args.instance
        self.app.session.post(uri, payload=dict(new_state="REBOOT",
                                                params={"reboot_type": parsed_args.type}))


class ListInstanceVolume(Lister):
    """List instance volume(s)."""

    def get_parser(self, prog_name):
        parser = super(ListInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes' % parsed_args.instance
        data = self.app.session.get(uri)

        return tabulate(data)


class ShowInstanceVolume(ShowOne):
    """Show instance volume details."""

    def get_parser(self, prog_name):
        parser = super(ShowInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes/%s' % (parsed_args.instance, parsed_args.volume)
        data = self.app.session.get(uri)

        return tabulate(data)


class AttachInstanceVolume(Command):
    """Attach instance volume."""

    def get_parser(self, prog_name):
        parser = super(AttachInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes' % parsed_args.instance
        payload = {'id': parsed_args.volume}

        self.app.session.post(uri, payload)


class DetachInstanceVolume(Command):
    """Detach instance volume."""

    def get_parser(self, prog_name):
        parser = super(DetachInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes/%s' %(parsed_args.instance, parsed_args.volume)
        self.app.session.delete(uri)


class ListInstanceInterface(Lister):
    """List instance interface(s)."""

    def get_parser(self, prog_name):
        parser = super(ListInstanceInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/interfaces' % parsed_args.instance
        data = self.app.session.get(uri)

        return tabulate(data)


class AttachInterface(Command):
    """Attach interface."""

    def get_parser(self, prog_name):
        parser = super(AttachInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--network',
            required=True,
            metavar='<network>',
            help='Network name or ID',

        )
        return parser

    def take_action(self, parsed_args):
        payload = {'net': parsed_args.network}
        uri = '/instances/%s/interfaces' % parsed_args.instance
        self.app.session.post(uri, payload)


class DetachInterface(Command):
    """Detach interface."""

    def get_parser(self, prog_name):
        parser = super(DetachInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--port-id',
            required=True,
            metavar='<port>',
            help='Port ID',

        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/interfaces/%s' %(parsed_args.instance, parsed_args.port_id)
        self.app.session.delete(uri)


class UpdateInstance(ShowOne):
    """Update instance name."""
    def get_parser(self, prog_name):
        parser = super(UpdateInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help=('Instance ID'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            required=True,
            help='New instance name.',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s' % parsed_args.instance
        payload = {
            "name": parsed_args.name
        }
        data = self.app.session.put(uri, payload)

        return tabulate(data)


class ListSecurityGroup(Lister):
    """List Instance Security Groups"""

    def get_parser(self, prog_name):
        parser = super(ListSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/security-groups' % parsed_args.instance
        data = self.app.session.get(uri)
        return tabulate(data)


class AddSecurityGroup(Command):
    """Add Security Group to instance"""

    def get_parser(self, prog_name):
        parser = super(AddSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--security-group',
            required=True,
            metavar='<security_group>',
            help='Security group name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        payload = {'security_group_ref': parsed_args.security_group}
        uri = '/instances/%s/security-groups' % parsed_args.instance
        self.app.session.post(uri, payload)


class RemoveSecurityGroup(Command):
    """Remove Security Group from Instance"""

    def get_parser(self, prog_name):
        parser = super(RemoveSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--security-group',
            required=True,
            metavar='<security_group>',
            help='Security group name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/security-groups/%s' % (parsed_args.instance,
                                                    parsed_args.security_group)
        self.app.session.delete(uri)



"""
    Common Utilitie functions.
"""
import sys


def tabulate(data):
    """Make dict or list to table ready tuples."""
    if isinstance(data, dict) and data:
        headers = data.keys()
        rows = [value for key, value in data.items()]
    elif isinstance(data, list) and data:
        headers = data[0].keys()
        rows = [[value for key, value in obj.items()] for obj in data]
    else:
        return (), ()

    return headers, rows


def is_tty():
    return sys.stdin.isatty()


def yes_or_no(confirmation_message, default_yes=False):
    if default_yes:
        options = "[Y/n]"
    else:
        options = "[y/N]"
    user_response = input(confirmation_message + options).lower()
    if user_response not in ('y', 'n'):
        return default_yes
    return user_response == 'y'



"""
    DigiCloud Security Service.
"""
from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate


class ListSecurityGroup(Lister):
    """List Security Groups."""

    def take_action(self, parsed_args):
        data = self.app.session.get('/security-groups')

        return tabulate(data)


class ShowSecurityGroup(ShowOne):
    """Show security group details."""

    def get_parser(self, prog_name):
        parser = super(ShowSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'security_group',
            metavar='<security_group>',
            help='Security group name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/%s' % parsed_args.security_group
        data = self.app.session.get(uri)

        return tabulate(data)


class CreateSecurityGroup(ShowOne):
    """Create Security Group."""

    def get_parser(self, prog_name):
        parser = super(CreateSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Security group name'
        )
        return parser

    def take_action(self, parsed_args):
        payload = {'name': parsed_args.name}
        data = self.app.session.post('/security-groups', payload)

        return tabulate(data)


class DeleteSecurityGroup(Command):
    """Delete Security Group."""

    def get_parser(self, prog_name):
        parser = super(DeleteSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'security_group',
            metavar='<security_group>',
            help='Security group name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/%s' % parsed_args.security_group
        self.app.session.delete(uri)


class UpdateSecurityGroup(Command):
    """Update Security Group."""

    def get_parser(self, prog_name):
        parser = super(UpdateSecurityGroup, self).get_parser(prog_name)
        parser.add_argument(
            'security_group',
            metavar='<security_group>',
            help='Security group name or ID'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Security group name'
        )
        return parser

    def take_action(self, parsed_args):
        uri = f'/security-groups/{parsed_args.security_group}'
        payload = {
            'name': parsed_args.name,
        }
        self.app.session.patch(uri, payload)


class ListSecurityGroupRule(Lister):
    """List Security Group Rules."""

    def get_parser(self, prog_name):
        parser = super(ListSecurityGroupRule, self).get_parser(prog_name)
        parser.add_argument(
            'security_group',
            metavar='<security_group>',
            help='Security group rule ID',
        )
        return parser

    def take_action(self, parsed_args):
        data = self.app.session.get('/security-groups/{}'.format(parsed_args.security_group))
        return tabulate(data['security_group_rules'])


class ShowSecurityGroupRule(ShowOne):
    """Show security group rule details."""

    def get_parser(self, prog_name):
        parser = super(ShowSecurityGroupRule, self).get_parser(prog_name)
        parser.add_argument(
            'security_group_rule',
            metavar='<security_group_rule>',
            help='Security group rule ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/_/rules/%s' % parsed_args.security_group_rule
        data = self.app.session.get(uri)

        return tabulate(data)


class CreateSecurityGroupRule(ShowOne):
    """Create Security Group Rule."""

    def get_parser(self, prog_name):
        parser = super(CreateSecurityGroupRule, self).get_parser(prog_name)
        parser.add_argument(
            'security_group',
            metavar='<security_group>',
            help='Security group name or ID'
        )
        parser.add_argument(
            '--direction',
            required=True,
            metavar='<direction>',
            help='Direction'
        )
        parser.add_argument(
            '--ethertype',
            default='IPv4',
            metavar='<ethertype>',
            help='IP version',
            choices=['IPv4', 'IPv6']
        )
        parser.add_argument(
            '--port-range-min',
            metavar='<port_range_min>',
            help='Minimum of port range'
        )
        parser.add_argument(
            '--port-range-max',
            metavar='<port_range_max>',
            help='Maximum of port range'
        )
        parser.add_argument('--protocol', required=True, metavar='<protocol>',
            help="Protocol")
        return parser

    def take_action(self, parsed_args):
        args = [
            'direction', 'ethertype', 'port_range_min',
            'port_range_max', 'protocol'
        ]
        payload = {arg: getattr(parsed_args, arg) for arg in args
                   if getattr(parsed_args, arg) is not None}
        uri = '/security-groups/%s/rules' % parsed_args.security_group
        data = self.app.session.post(uri, payload)

        return tabulate(data)


class DeleteSecurityGroupRule(Command):
    """Delete Security Group Rule."""

    def get_parser(self, prog_name):
        parser = super(DeleteSecurityGroupRule, self).get_parser(prog_name)
        parser.add_argument(
            'security_group_rule',
            metavar='<security_group_rule>',
            help='Security group rule ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/_/rules/%s' % parsed_args.security_group_rule
        self.app.session.delete(uri)



"""
    DigiCloud Volume Service.
"""
from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..utils import tabulate, is_tty


class ListVolume(Lister):
    """List volumes."""

    def take_action(self, parsed_args):
        data = self.app.session.get('/volumes')

        return tabulate(data)


class ShowVolume(ShowOne):
    """Show volume details."""

    def get_parser(self, prog_name):
        parser = super(ShowVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/volumes/%s' % parsed_args.volume
        data = self.app.session.get(uri)

        return tabulate(data)


class DeleteVolume(Command):
    """Delete volume."""

    def get_parser(self, prog_name):
        parser = super(DeleteVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID'
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        if not self.confirm(parsed_args):
            return
        uri = '/volumes/%s' % parsed_args.volume
        self.app.session.delete(uri)

    def confirm(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            volume = self.app.session.get('/volumes/%s' % parsed_args.volume)
            user_response = input(
                "You're about to delete a volume named {} with {}GB size. "
                "Are you sure? [yes/no]".format(
                    volume['name'],
                    volume['size']
                ))
            if user_response == "yes":
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write("Unable to perform 'delete volume' operation in non-interactive mode,"
                                  " without '--i-am-sure' switch\n")
            return False


class CreateVolume(ShowOne):
    """Create Volume."""

    def get_parser(self, prog_name):
        parser = super(CreateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Volume name'
        )
        parser.add_argument(
            '--size',
            required=True,
            metavar='<size>',
            help='Volume size'
        )

        parser.add_argument(
            '--type',
            required=True,
            metavar='<type>',
            help='Volume type, could be SSD or HDD',
            choices=['SSD', 'HDD']
        )

        return parser

    def take_action(self, parsed_args):
        volume_type = parsed_args.type.upper()
        if volume_type not in {"SSD", "HDD"}:
            raise Exception("--type parameter should be either SSD or HDD")
        payload = {
            'name': parsed_args.name,
            'size': parsed_args.size,
            'volume_type': volume_type
        }
        data = self.app.session.post('/volumes', payload)

        return tabulate(data)


class UpdateVolume(ShowOne):
    """Update volume."""
    def get_parser(self, prog_name):
        parser = super(UpdateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=('Volume ID'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            required=True,
            help=('Volume new name'),
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/volumes/%s' % parsed_args.volume
        payload = {
            "name": parsed_args.name
        }
        data = self.app.session.patch(uri, payload)

        return tabulate(data)



"""
    Configure ``setup.py`` for digicloud cli app.
"""
from setuptools import setup, find_packages


try:
    LONG_DESCRIPTION = open('README.md', 'rt').read()
except IOError:
    LONG_DESCRIPTION = ''

setup(
    name='digicloud',
    version='1.1.4',
    description='DigiCloud CLI',

    long_description=LONG_DESCRIPTION,

    author='Arash Shams',
    author_email='a.shams@digikala.com',
    url='https://digicloud.ir',
    download_url='',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Environment :: Console',
    ],
    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=[
        'cliff',
        'requests',
        'setuptools',
    ],
    extras_require={
        'dev': [
            'pytest==6.1.0'
        ]
    },
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'digicloud = digicloud.main:main'
        ],
        'digicloud.cli': [
            'account_login = digicloud.commands.account:Login',
            'account_logout = digicloud.commands.account:Logout',
            'namespace_current = digicloud.commands.namespace:CurrentNamespace',
            'namespace_select = digicloud.commands.namespace:SelectNamespace',
            'namespace_create = digicloud.commands.namespace:CreateNamespace',
            'namespace_delete = digicloud.commands.namespace:DeleteNamespace',
            'namespace_leave = digicloud.commands.namespace:LeaveNamespace',
            'namespace_list = digicloud.commands.namespace:ListNamespace',
            'namespace_show = digicloud.commands.namespace:ShowNamespace',
            'namespace_member_list = digicloud.commands.namespace:ListNamespaceMember',
            'namespace_member_add = digicloud.commands.namespace:InviteMember',
            'namespace_invitation_accept = digicloud.commands.namespace:AcceptInvitation',
            'namespace_invitation_list = digicloud.commands.namespace:ListInvitation',
            'namespace_member_delete = digicloud.commands.namespace:DeleteNamespaceMember',
            'instance_type_list = digicloud.commands.instance_type:ListInstanceType',
            'instance_type_show = digicloud.commands.instance_type:ShowInstanceType',
            'floating_ip_list = digicloud.commands.access:ListFloatingIp',
            'floating_ip_show = digicloud.commands.access:ShowFloatingIp',
            'floating_ip_create = digicloud.commands.access:CreateFloatingIp',
            'floating_ip_delete = digicloud.commands.access:DeleteFloatingIp',
            'floating_ip_associate = digicloud.commands.access:AssociateFloatingIp',
            'floating_ip_revoke = digicloud.commands.access:RevokeFloatingIp',
            'image_list = digicloud.commands.image:ListImage',
            'image_show = digicloud.commands.image:ShowImage',
            'instance_list = digicloud.commands.instance:ListInstance',
            'instance_create = digicloud.commands.instance:CreateInstance',
            'instance_delete = digicloud.commands.instance:DeleteInstance',
            'instance_update = digicloud.commands.instance:UpdateInstance',
            'instance_start = digicloud.commands.instance:StartInstance',
            'instance_stop = digicloud.commands.instance:StopInstance',
            'instance_suspend = digicloud.commands.instance:SuspendInstance',
            'instance_resume = digicloud.commands.instance:ResumeInstance',
            'instance_reboot = digicloud.commands.instance:RebootInstance',
            'instance_show = digicloud.commands.instance:ShowInstance',
            'instance_volume_list = digicloud.commands.instance:ListInstanceVolume',
            'instance_volume_show = digicloud.commands.instance:ShowInstanceVolume',
            'instance_volume_attach = digicloud.commands.instance:AttachInstanceVolume',
            'instance_volume_detach = digicloud.commands.instance:DetachInstanceVolume',
            'instance_interface_list = digicloud.commands.instance:ListInstanceInterface',
            'instance_interface_attach = digicloud.commands.instance:AttachInterface',
            'instance_interface_detach = digicloud.commands.instance:DetachInterface',
            'instance_security_group_list = digicloud.commands.instance:ListSecurityGroup',
            'instance_security_group_add = digicloud.commands.instance:AddSecurityGroup',
            'instance_security_group_remove = digicloud.commands.instance:RemoveSecurityGroup',
            'volume_list = digicloud.commands.volume:ListVolume',
            'volume_show = digicloud.commands.volume:ShowVolume',
            'volume_delete = digicloud.commands.volume:DeleteVolume',
            'volume_create = digicloud.commands.volume:CreateVolume',
            'volume_update = digicloud.commands.volume:UpdateVolume',
            'network_list = digicloud.commands.network:ListNetwork',
            'network_show = digicloud.commands.network:ShowNetwork',
            'network_delete = digicloud.commands.network:DeleteNetwork',
            'network_create = digicloud.commands.network:CreateNetwork',
            'network_update = digicloud.commands.network:UpdateNetwork',
            'network_port_list = digicloud.commands.network:ListNetworkPorts',
            'security_group_list = digicloud.commands.security:ListSecurityGroup',
            'security_group_show = digicloud.commands.security:ShowSecurityGroup',
            'security_group_create = digicloud.commands.security:CreateSecurityGroup',
            'security_group_delete = digicloud.commands.security:DeleteSecurityGroup',
            'security_group_update = digicloud.commands.security:UpdateSecurityGroup',
            'security_group_rule_list = digicloud.commands.security:ListSecurityGroupRule',
            'security_group_rule_show = digicloud.commands.security:ShowSecurityGroupRule',
            'security_group_rule_create = digicloud.commands.security:CreateSecurityGroupRule',
            'security_group_rule_delete = digicloud.commands.security:DeleteSecurityGroupRule',
            'subnet_create = digicloud.commands.subnet:CreateSubnet',
            'subnet_list = digicloud.commands.subnet:ListSubnet',
            'subnet_show = digicloud.commands.subnet:ShowSubnet',
            'subnet_delete = digicloud.commands.subnet:DeleteSubnet',
            'subnet_update = digicloud.commands.subnet:UpdateSubnet',
            'ssh_key_list = digicloud.commands.ssh_key:ListSSHKey',
            'ssh_key_show = digicloud.commands.ssh_key:ShowSSHKey',
            'ssh_key_delete = digicloud.commands.ssh_key:DeleteSSHKey',
            'ssh_key_create = digicloud.commands.ssh_key:CreateSSHKey',
            'router_list = digicloud.commands.router:ListRouter',
            'router_show = digicloud.commands.router:ShowRouter',
            'router_update = digicloud.commands.router:UpdateRouter',
            'router_delete = digicloud.commands.router:DeleteRouter',
            'router_create = digicloud.commands.router:CreateRouter',
            'router_interface_list = digicloud.commands.router:ListRouterInterface',
            'router_interface_add = digicloud.commands.router:AddRouterInterface',
            'router_interface_remove = digicloud.commands.router:RemoveRouterInterface',
            'router_external_add = digicloud.commands.router:AddRouterExternal',
            'router_external_remove = digicloud.commands.router:RemoveRouterExternal',
            'router_static_list = digicloud.commands.router:ListRouterStatic',
            'router_static_add = digicloud.commands.router:AddRouterStatic',
            'router_static_delete = digicloud.commands.router:DeleteRouterStatic',
            'region_list = digicloud.commands.region:ListRegions',
            'region_select = digicloud.commands.region:SelectRegion',
            'region_current = digicloud.commands.region:CurrentRegion',
        ],
    },
    zip_safe=False,
)



"""
    DigiCloud Subnet Service.
"""

from cliff.command import Command
from cliff.lister import Lister
from cliff.show import ShowOne

from ..cli import parseractions
from ..utils import tabulate


class CreateSubnet(ShowOne):
    """Create Network Subnet"""

    def get_parser(self, prog_name):
        parser = super(CreateSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Subnet name'
        )
        parser.add_argument(
            '--ip-version',
            default='4',
            metavar='<ip_version>',
            help='IP version, default to IPv4'
        )
        parser.add_argument(
            '--cidr',
            required=True,
            metavar='<cidr>',
            help='CIDR, e.g: 10.0.0.0/16'
        )
        parser.add_argument(
            '--network',
            required=True,
            metavar='<network>',
            help='network name or ID'
        )
        parser.add_argument(
            '--enable-dhcp',
            metavar='<enable_dhcp>',
            help='Enable DHCP for this subnet'
        )
        parser.add_argument(
            '--allocation-pool',
            metavar='start=<ip-address>,end=<ip-address>',
            dest='allocation_pools',
            action=parseractions.MultiKeyValueAction,
            required_keys=['start', 'end'],
            help='Allocation pool IP addresses for this subnet e.g.: start=192.168.199.2,'
                 'end=192.168.199.254 (repeat option to add multiple IP addresses)'
        )
        parser.add_argument(
            '--dns-server',
            metavar='<dns_server>',
            action='append',
            dest='dns_server',
            help="DNS server for this subnet (repeat option to set multiple DNS servers)"
        )
        parser.add_argument(
            '--host-route',
            metavar='destination=<subnet>,nexthop=<ip-address>',
            dest='host_route',
            action=parseractions.MultiKeyValueAction,
            required_keys=['destination', 'nexthop'],
            help="Additional route for this subnet "
                 "e.g.: destination=10.10.0.0/16,nexthop=192.168.71.254 "
                 "destination: destination subnet (in CIDR notation) "
                 "nexthop: nexthop IP address "
                 "(repeat option to add multiple routes)")

        return parser

    def take_action(self, parsed_args):
        uri = '/networks/%s/subnets' % parsed_args.network
        payload = {
            'ip_version': parsed_args.ip_version,
            'cidr': parsed_args.cidr,
            'subnet_name': parsed_args.name,
            **self._get_optional_fields(parsed_args)
        }
        data = self.app.session.post(uri, payload)

        return tabulate(data)

    def _get_optional_fields(self, args):
        optional_args = {}

        if args.enable_dhcp is not None:
            optional_args["enable_dhcp"] = args.enable_dhcp

        if args.allocation_pools is not None:
            optional_args["allocation_pools"] = args.allocation_pools

        if args.dns_server is not None:
            optional_args["dns_servers"] = args.dns_server

        if args.host_route is not None:
            optional_args["host_routes"] = args.host_route

        return optional_args


class ListSubnet(Lister):
    """List subnets."""

    def take_action(self, parsed_args):
        data = self.app.session.get('/subnets')

        return tabulate(data)


class ShowSubnet(ShowOne):
    """Show subnet details."""

    def get_parser(self, prog_name):
        parser = super(ShowSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='Subnet name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        data = self.app.session.get(uri)

        return tabulate(data)


class DeleteSubnet(Command):
    """Delete subnet."""

    def get_parser(self, prog_name):
        parser = super(DeleteSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='subnet name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        self.app.session.delete(uri)


class UpdateSubnet(ShowOne):
    """Update subnets."""
    def get_parser(self, prog_name):
        parser = super(UpdateSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='Subnet ID',
        )
        parser.add_argument(
            '--name',
            required=True,
            metavar='<Name>',
            help='New name for subnet.'
        )
        parser.add_argument(
            '--description',
            metavar='<Description>',
            help='New description for subnet.'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        payload = {
            'name': parsed_args.name,
            'description': parsed_args.description,
        }
        data = self.app.session.patch(uri, payload)
        return tabulate(data)



"""
    DigiCloud Image Service
"""
from cliff.show import ShowOne
from cliff.lister import Lister

from digicloud import formatter
from ..utils import tabulate


class ListImage(Lister):
    """List images"""

    def take_action(self, parsed_args):
        data = self.app.session.get('/images')

        for image in data:
            image['size'] = formatter.HumanBytes.format(image['size'], True)

        return tabulate(data)


class ShowImage(ShowOne):
    """Show image details."""

    def get_parser(self, prog_name):
        parser = super(ShowImage, self).get_parser(prog_name)
        parser.add_argument(
            'image',
            metavar='<image>',
            help='Image name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/images/%s' % parsed_args.image
        data = self.app.session.get(uri)

        data['size'] = formatter.HumanBytes.format(data['size'], True)

        return tabulate(data)
